/*
 * dll_wcsnrtombs_locale ()
 * for conversion from UCS-4 to multibyte character
 * programed
 *   by A. Yoshiyama (yosshy@jedi.seg.kobe-u.ac.jp)
 *   on Dec.17,1996
 */

#include <wchar.h>
#include <wcsmbs_locale.h>
#include <ucs2_2_table.h>
#include <errno.h>

#ifndef EILSEQ
#define EILSEQ EINVAL
#endif

static mbstate_t internal;

size_t
dll_wcsnrtombs_locale (dst, src, nwc, len, ps)
     char *dst;
     const wchar_t **src;
     size_t nwc;
     size_t len;
     mbstate_t *ps;
{
  size_t written = 0;
  const wchar_t *run;
  int codetype;

#ifdef HAS_ESCAPE_SEQUENCE
  int codetype_last;
#endif

  if (ps == NULL)
    ps = &internal;

#ifdef HAS_ESCAPE_SEQUENCE
  codetype_last = ps->count;
#endif

  if (src == NULL || *src == NULL)
#ifdef HAS_ESCAPE_SEQUENCE
    {
      ps->count = 0;
      return (size_t)codetype_last;
    }
#else
    return (size_t)0;
#endif

  if (dst == NULL)
    len = ~0;

  for (run = *src; written < len && nwc-- > 0; run++)
    {
      if (*run < 0 || *run > 0x7fffffff)
	{
	  __set_errno (EILSEQ);
	  return (size_t) -1;
	}
      if (*run == L'\0')
	{
	  /* Found the end.  */
	  if (dst != NULL)
	    *dst = '\0';
	  *src = NULL;
	  return (size_t)written;
	}
      else
	{
	  int wc_seek;
	  MCKEY_T key_tmp, key_last, wc_key;

	  key_last = 0;
	  wc_key = (MCKEY_T)((*run & 0xFC00) >> 10);

	  for (wc_seek = ucs2_2_table1[(int)*run & 0x3FF];
	       (key_tmp = (ucs2_2_table2_key[wc_seek] & 0x3F)) >= key_last;
	       wc_seek++)
	    {
	      key_last = key_tmp;

	      if (wc_key == key_tmp)
		{
		  int seek;
		  codetype = (unsigned int)ucs2_2_table2_key[wc_seek] >> 6;
#ifdef HAS_ESCAPE_SEQUENCE
		  if (codetype != codetype_last)
		    {
		      codetype_last = codetype;
		      if (written + mbc_size[codetype] + es_size[codetype] > len)
			goto break_while;
		      if (dst != NULL)
			for (seek=0; seek < es_size[codetype]; seek++)
			  *dst++ = es[codetype][seek];
		      written += es_size[codetype];
		    }
		  else 
		    if (written + mbc_size[codetype] > len)
		      goto break_while;
#else
		  if (written + mbc_size[codetype] + es_size[codetype] > len)
		    goto break_while;
		  if (dst != NULL) 
		    for (seek=0; seek < es_size[codetype]; seek++)
		      *dst++ = es[codetype][seek];
		  written += es_size[codetype];
#endif
		  if (dst != NULL)
		    switch (mbc_size[codetype])
		      {
		      case 2:
			*dst++ = (char)((0xFF00 & ucs2_2_table2[wc_seek]) >> 8);
			*dst++ = (char) (0x00FF & ucs2_2_table2[wc_seek]);
			break;
		      case 1:
			*dst++ = (char)(0xFF & ucs2_2_table2[wc_seek]);
			break;
		      default:
			__set_errno(EINVAL);
			return (size_t)-2;
		      }
		  written += mbc_size[codetype];
		}  /* end of "if (wc_key == key_tmp)" */
	    }  /* end of "for (wc_seek ...)" */
	}  /* end of "if (wc == L'\0') ... else" */
    }  /* end of "while (written < len && nwc -- > 0)" */

 break_while:

  *src = run;

#ifdef HAS_ESCAPE_SEQUENCE
  ps->count = codetype_last;
#endif

  return (size_t)written;
}
